#define WIN32_LEAN_AND_MEAN
#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <string.h>

double reltime() {
  static LONGLONG pt=0;
  static double prec;
  static BOOL supp;
  extern int tmcnt;
  static tmcnt1=0;
  LARGE_INTEGER ptime;
  double d;
  if (pt==0) {
	LARGE_INTEGER ftime;
	supp=QueryPerformanceFrequency(&ftime);
	printf("TimerFreq. %li\n",ftime);
	QueryPerformanceCounter(&ptime);
	prec=1./ftime.QuadPart;
	pt=ptime.QuadPart;
	return 0;
  }
  QueryPerformanceCounter(&ptime);
  d=(ptime.QuadPart-pt)*prec;
  pt=ptime.QuadPart;
  if (d<0) {
	printf("Program error\n");
	return 0;
  }
  return d;
}

int main(int argc,char *argv[]) {
  DWORD tmb,n;
  if (argc<3) {
	fprintf(stderr,"This program tests memory read & write speed. v0.4\n");
	fprintf(stderr,"Usage: %s number_of_MBytes number_of_iterations\n",argv[0]);
	fprintf(stderr,"Example: testmem 10 10\n",argv[0]);
	return 1;
  }
  DWORD numMB,numTimes;
  if (sscanf(argv[1],"%i",&numMB)<1) {
	fprintf(stderr,"Parameter 1 incorrect.\n");
	return 2;
  }
  if (sscanf(argv[2],"%i",&numTimes)<1){
	fprintf(stderr,"Parameter 2 incorrect.\n");
	return 3;
  }
  if (numTimes*numMB>=30000) {
	fprintf(stderr,"This can take a long time. Press ^C to abort.\n");
  }
  tmb=numMB*1024*1024;
  printf("Writing %u MB %u times\n",numMB,numTimes);
  DWORD *c=(DWORD*)malloc(tmb);
  if (!c) {
	fprintf(stderr,"Program error: out of memory\n");
	return 4;
  }
  reltime();
  //Commit pages:
  for(n=0;n<tmb/4;n+=512) c[n]=0x01010101;
  for(n=0;n<tmb/4;n+=512) c[n]=0x02020202;

  SetPriorityClass(GetCurrentProcess(),REALTIME_PRIORITY_CLASS);
  /// ====================== write speed test ========================
  double time;

  //fprintf(stderr,"Now testing 64-bit writing (with FPU). Press ^C to abort\n");
  reltime();
  for(n=0;n<numTimes;n++) {
    __asm fld time
    __asm mov ecx,tmb __asm mov eax,n __asm mov esi,c __asm shr ecx,3
wfl:__asm fst qword ptr [esi] __asm add esi,8 __asm dec ecx __asm jnz wfl
    __asm fstp st
  }
  time=reltime();
  printf("64-bit writing took %lf seconds, thus %u KB/s, %uK FP writes/s\n",
     	time,int(numMB*numTimes*1024/time),int(numMB*numTimes*1024/time)/8);

  __asm mov eax,1 __asm cpuid __asm mov n,edx
  if (n&0x800000) {
	//fprintf(stderr,"Now testing 64-bit writing (with MMX).\n");
	reltime();
	for(n=0;n<numTimes;n++) {
	  __asm movq mm0,time
	  __asm mov ecx,tmb __asm mov eax,n __asm mov esi,c __asm shr ecx,3
wmmx: __asm MOVQ [esi],mm0 __asm add esi,8 __asm dec ecx __asm jnz wmmx
	  __asm EMMS
	}
	time=reltime();
	printf("64-bit writing took %lf seconds, thus %u KB/s, %uK MMX writes/s\n",
    	time,int(numMB*numTimes*1024/time),int(numMB*numTimes*1024/time)/8);
  }
  time=reltime();
  for(n=0;n<numTimes;n++) {
	__asm mov ecx,tmb __asm mov eax,n __asm mov edi,c __asm shr ecx,2
	__asm cld __asm rep stosd
  }
  time=reltime();
  printf("32-bit writing took %lf seconds, thus %u KB/s, %uK writes/s\n",
  		time,int(numMB*numTimes*1024/time),int(numMB*numTimes*1024/time)/4);
  time=reltime();
  for(n=0;n<numTimes;n++) {
	__asm mov ecx,tmb __asm mov eax,n __asm mov edi,c __asm shr ecx,1
	__asm cld __asm rep stosw
  }
  time=reltime();
  printf("16-bit writing took %lf seconds, thus %u KB/s, %uK writes/s\n",
  		time,int(numMB*numTimes*1024/time),int(numMB*numTimes*1024/time)/2);
  time=reltime();
  for(n=0;n<numTimes;n++) {
	__asm mov ecx,tmb __asm mov eax,n __asm mov edi,c //__asm shr ecx,2
	__asm cld __asm rep stosb
  }
  time=reltime();
  printf(" 8-bit writing took %lf seconds, thus %u KB/s, %uK writes/s\n",
  		time,int(numMB*numTimes*1024/time),int(numMB*numTimes*1024/time)/1);
  /// ======================= read speed test ========================

  printf("\n"); Sleep(200);
  //prepare array:
    __asm fldz __asm mov ecx,tmb __asm mov esi,c __asm shr ecx,3
pfl:__asm fst qword ptr [esi] __asm add esi,8 __asm dec ecx __asm jnz pfl __asm fstp st

  reltime();
  for(n=0;n<numTimes;n++) {
    __asm mov ecx,tmb __asm mov esi,c __asm shr ecx,3 __asm fld1
rfl:__asm fadd qword ptr [esi] __asm add esi,8 __asm dec ecx //__asm fstp st
    __asm jnz rfl __asm fstp st
  }
  time=reltime();
  printf("64-bit reading took %lf seconds, thus %u KB/s, %uK FP reads/s\n",
     	time,int(numMB*numTimes*1024/time),int(numMB*numTimes*1024/time)/8);

  __asm mov eax,1 __asm cpuid __asm mov n,edx
  if (n&0x800000) {
	reltime();
	for(n=0;n<numTimes;n++) {
	  __asm mov ecx,tmb __asm mov eax,n __asm mov esi,c __asm shr ecx,3
rmmx: __asm movq mm0,[esi] __asm add esi,8 __asm dec ecx __asm jnz rmmx
	  __asm EMMS
	}
	time=reltime();
	printf("64-bit reading took %lf seconds, thus %u KB/s, %uK MMX reads/s\n",
    	time,int(numMB*numTimes*1024/time),int(numMB*numTimes*1024/time)/8);
  }
  time=reltime();
  for(n=0;n<numTimes;n++) {
	__asm mov ecx,tmb __asm mov edi,c __asm shr ecx,2
r32:__asm mov eax,[edi] __asm add edi,4 __asm dec ecx __asm jnz r32
  }
  time=reltime();
  printf("32-bit reading took %lf seconds, thus %u KB/s, %uK reads/s\n",
  		time,int(numMB*numTimes*1024/time),int(numMB*numTimes*1024/time)/4);

  time=reltime();
  for(n=0;n<numTimes;n++) {
	__asm mov ecx,tmb __asm mov eax,n __asm mov edi,c __asm shr ecx,1
r16:__asm movzx eax,word ptr [edi] __asm add edi,2 __asm dec ecx __asm jnz r16
  }
  time=reltime();
  printf("16-bit reading took %lf seconds, thus %u KB/s, %uK reads/s\n",
  		time,int(numMB*numTimes*1024/time),int(numMB*numTimes*1024/time)/2);
  time=reltime();
  for(n=0;n<numTimes;n++) {
	__asm mov ecx,tmb __asm mov eax,n __asm mov edi,c
r8: __asm mov al,[edi] __asm inc edi __asm dec ecx __asm jnz r8
  }
  time=reltime();
  printf(" 8-bit reading took %lf seconds, thus %u KB/s, %uK reads/s\n",
  		time,int(numMB*numTimes*1024/time),int(numMB*numTimes*1024/time)/1);

  SetPriorityClass(GetCurrentProcess(),NORMAL_PRIORITY_CLASS);
  free(c);
  return 0;
}